import os
import abc
import redis
import typing
import contextlib
import dataclasses


class Nested(abc.ABC):

    # Instance globals
    _key: str = None  # type: ignore
    _redis: redis.Redis = None  # type: ignore
    _master: str = None  # type: ignore

    # Type globals
    ENCODING: str = "utf-8"

    def __init__(self, key: str, redis: redis.Redis, master: typing.Optional[str] = None) -> None:
        # Set internal input parameters
        self._key = key
        self._redis = redis
        self._master = master or key

    @abc.abstractmethod
    def initialize(self, value: typing.Any) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def deinitialize(self) -> None:
        raise NotImplementedError()

    def _fetch_by_identifier(self, identifier: typing.Union[str, bytes]) -> typing.Any:
        # Make sure the identifier is a string
        if not isinstance(identifier, str):
            identifier = identifier.decode(self.ENCODING)

        # Split identifier to item type and encoded value
        redis_identifier, encoded_item_value = identifier.split(":", 1)

        # Parse item type and encoded value to redis identifier and decoded value
        decoded_item_value = self._decode(encoded_item_value)

        # Check if the item is nested
        for nested_type in NESTED_TYPES:
            # Check whether the item type matches the redis identifier
            if redis_identifier != nested_type.redis_identifier:
                continue

            # Found a nested type match, create a nested instance
            return nested_type.nested_class(key=decoded_item_value, redis=self._redis, master=self._master)

        # Return the decoded value
        return decoded_item_value

    def _delete_by_identifier(self, identifier: typing.Union[str, bytes]) -> None:
        # Make sure the identifier is a string
        if not isinstance(identifier, str):
            identifier = identifier.decode(self.ENCODING)

        # Fetch the nested object
        nested_item = self._fetch_by_identifier(identifier)

        # If the item is nested, deconstruct it
        if not isinstance(nested_item, Nested):
            return

        # Deinitialize the nested item
        nested_item.deinitialize()

    @contextlib.contextmanager
    def _create_identifier_from_value(self, value: typing.Any) -> typing.Iterator[str]:
        # Check whether value supports nesting
        for nested_type in NESTED_TYPES:
            # Check if the value is an instance of the current type
            if not isinstance(value, nested_type.conversion_type):
                continue

            # Create a new nested name
            nested_name = f"{self._master}:{os.urandom(10).hex()}"

            # Create a new nested class instance
            nested_instance = nested_type.nested_class(key=nested_name, redis=self._redis, master=self._master)
            nested_instance.initialize(value)

            try:
                # Yield the nested identifier
                yield f"{nested_type.redis_identifier}:{self._encode(nested_name)}"
            except:
                # If there was a failure to insert the identifier, delete the nested item
                self._redis.delete(nested_name)

                # Re-raise the exception
                raise

            # Nothing more to do
            return

        # Yield the regular value
        yield f":{self._encode(value)}"

    def _encode(self, object: typing.Any) -> str:
        # Return the representation of the object
        return repr(object)

    def _decode(self, value: typing.Union[str, bytes]) -> typing.Any:
        # Make sure the value is a string
        if not isinstance(value, str):
            value = value.decode(self.ENCODING)

        # Evaluate the value
        return eval(value)


@dataclasses.dataclass
class NestedType(object):

    # Database identifier
    redis_identifier: str

    # Nested class
    nested_class: typing.Type[Nested]

    # Conversion type
    conversion_type: typing.Type[typing.Collection[typing.Any]]


# List of all supported nested types
NESTED_TYPES: typing.List[NestedType] = list()



import typing
import contextlib

# Import abstract types
from collections.abc import Mapping, Iterable

# Import extension objects
from rednest.nested import Nested, NestedType, NESTED_TYPES

# Create default object so that None can be used as default value
DEFAULT = object()


class Dictionary(typing.MutableMapping[typing.Any, typing.Any], Nested):

    # Bunch mode switch
    BUNCH = True

    def initialize(self, value: typing.Dict[typing.Any, typing.Any]) -> None:
        # De-initialize before initializing
        self.deinitialize()

        # Update the dictionary
        self.update(value)

    def deinitialize(self) -> None:
        # Clear the dictionary
        self.clear()

    def _identifier_from_key(self, key: typing.Any) -> typing.Union[str, bytes]:
        # Fetch the identifier from the hash
        identifier = self._redis.hget(self._key, self._encode(key))

        # If the response is empty, item does not exist
        if identifier is None:
            raise KeyError(key)

        # Make sure identifier is a string or bytes
        if not isinstance(identifier, (str, bytes)):
            raise TypeError(identifier)

        # Return the identifier
        return identifier

    def __getitem__(self, key: typing.Any) -> typing.Any:
        # Fetch the identifier, then return the value
        return self._fetch_by_identifier(self._identifier_from_key(key))

    def __setitem__(self, key: typing.Any, value: typing.Any) -> None:
        # Fetch the identifier
        original_identifier = self._redis.hget(self._key, self._encode(key))

        # Insert a new value
        with self._create_identifier_from_value(value) as identifier:
            self._redis.hset(self._key, self._encode(key), identifier)

        # If original identifier is not defined, there is nothing to delete
        if original_identifier is None:
            return

        # Make sure the original identifier is a string or bytes
        if not isinstance(original_identifier, (str, bytes)):
            raise TypeError(original_identifier)

        # Delete the original nested value
        self._delete_by_identifier(original_identifier)

    def __delitem__(self, key: typing.Any) -> None:
        # Fetch the identifier
        identifier = self._identifier_from_key(key)

        # Delete the key from hash
        self._redis.hdel(self._key, self._encode(key))

        # Delete the nested value
        self._delete_by_identifier(identifier)

    def __contains__(self, key: typing.Any) -> bool:
        # Make sure key exists in database
        return bool(self._redis.hexists(self._key, self._encode(key)))

    def __iter__(self) -> typing.Iterator[typing.Any]:
        # Fetch all hash keys
        encoded_keys = self._redis.hkeys(self._key)

        # Make sure encoded keys is iterable
        if not isinstance(encoded_keys, Iterable):
            raise TypeError(encoded_keys)

        # Loop over all object keys
        for encoded_key in encoded_keys:
            # Make sure the encoded key is a string or bytes
            if not isinstance(encoded_key, (str, bytes)):
                raise TypeError(encoded_key)

            # Yield the decoded key
            yield self._decode(encoded_key)

    def __len__(self) -> int:
        # Fetch the length of the hash
        length = self._redis.hlen(self._key)

        # Make sure the length is an integer
        if not isinstance(length, int):
            raise TypeError(length)

        # Return the hash length
        return length

    def __repr__(self) -> str:
        # Format the data like a dictionary
        return "{%s}" % ", ".join("%r: %r" % item for item in self.items())

    def __eq__(self, other: typing.Any) -> bool:
        # Make sure the other object is a mapping
        if not isinstance(other, Mapping):
            return False

        # Make sure all keys exist
        if set(self.keys()) != set(other.keys()):
            return False

        # Loop over all keys
        for key in self:
            # Check whether the value equals
            if self[key] != other[key]:
                return False

        # Comparison succeeded
        return True

    def pop(self, key: typing.Any, default: typing.Any = DEFAULT) -> typing.Any:
        try:
            # Fetch the original value
            value = self[key]

            # Try copying the value
            with contextlib.suppress(AttributeError):
                value = value.copy()

            # Delete the item
            del self[key]

            # Return the value
            return value
        except KeyError:
            # Check if a default is defined
            if default != DEFAULT:
                return default

            # Reraise exception
            raise

    def popitem(self) -> typing.Tuple[typing.Any, typing.Any]:
        # Convert self to list
        keys = list(self)

        # If the list is empty, raise
        if not keys:
            raise KeyError()

        # Pop a key from the list
        key = keys.pop()

        # Return the key and the value
        return key, self.pop(key)

    def copy(self) -> typing.Dict[typing.Any, typing.Any]:
        # Create initial bunch
        output = dict()

        # Loop over keys
        for key in self:
            # Fetch value of key
            value = self[key]

            # Try copying the value
            with contextlib.suppress(AttributeError):
                value = value.copy()

            # Update the bunch
            output[key] = value

        # Return the created output
        return output

    def setdefaults(self, *dictionaries: typing.Dict[typing.Any, typing.Any], **values: typing.Dict[typing.Any, typing.Any]) -> None:
        # Update values to include all dicts
        for dictionary in dictionaries:
            values.update(dictionary)

        # Loop over all items and set the default value
        for key, value in values.items():
            self.setdefault(key, value)

    # If bunch mode is enabled (on by default, define some more functions)
    if BUNCH:

        def __getattr__(self, key: str) -> typing.Any:
            try:
                return object.__getattribute__(self, key)
            except AttributeError:
                # Key is not in prototype chain, try returning
                try:
                    return self[key]
                except KeyError:
                    # Replace KeyErrors with AttributeErrors
                    raise AttributeError(key)

        def __setattr__(self, key: str, value: typing.Any) -> None:
            try:
                object.__getattribute__(self, key)
            except AttributeError:
                # Set the item
                self[key] = value
            else:
                # Key is in prototype chain, set it
                object.__setattr__(self, key, value)

        def __delattr__(self, key: str) -> None:
            try:
                object.__getattribute__(self, key)
            except AttributeError:
                # Delete the item
                try:
                    del self[key]
                except KeyError:
                    # Replace KeyErrors with AttributeErrors
                    raise AttributeError(key)
            else:
                # Key is in prototype chain, delete it
                object.__delattr__(self, key)


# Register nested object
NESTED_TYPES.append(NestedType("hash", Dictionary, dict))



import os
import typing
import contextlib

# Import abstract types
from collections.abc import Sequence, Iterable

# Import extension objects
from rednest.nested import Nested, NestedType, NESTED_TYPES


class List(typing.MutableSequence[typing.Any], Nested):

    def initialize(self, value: typing.List[typing.Any]) -> None:
        # De-initialize before initializing
        self.deinitialize()

        # Update the list
        self[:] = value

    def deinitialize(self) -> None:
        # Clear the list
        self.clear()

    def _identifier_from_index(self, index: int) -> typing.Union[str, bytes]:
        # Request the list at slice index->index
        identifier_response = self._redis.lrange(self._key, index, index)

        # If the response is empty, item does not exist
        if not identifier_response:
            raise IndexError(index)

        # Make sure the identifier response is a list
        if not isinstance(identifier_response, Iterable):
            raise TypeError(identifier_response)

        # Fetch the only item in the response
        identifier, = identifier_response

        # Make sure the identifier is a string or bytes
        if not isinstance(identifier, (str, bytes)):
            raise TypeError(identifier)

        # Return the identifier
        return identifier

    def __getitem__(self, index: typing.Union[int, slice]) -> typing.Union[typing.Any, typing.List[typing.Any]]:
        # If a slice is provided, return a list of items
        if isinstance(index, slice):
            # Create a list with all of the requested items
            return [self[item_index] for item_index in range(*index.indices(len(self)))]

        # If index is negative, add the length to it
        if index < 0:
            index += len(self)

        # Fetch the identifier
        identifier = self._identifier_from_index(index)

        # Return the nested value
        return self._fetch_by_identifier(identifier)

    def __setitem__(self, index: typing.Union[int, slice], value: typing.Union[typing.Any, typing.Sequence[typing.Any]]) -> None:
        # If a slice is provided, splice the list
        if isinstance(index, slice):
            # Fetch slice parameters
            start, stop, step = index.indices(len(self))

            # If the step is more then 1, the value length must match the indice count
            if step > 1 and len(value) != len(range(start, stop, step)):
                raise ValueError(f"Attempted to assign sequence of incompatible size {len(value)}")

            # Create an iterator for the value
            iterator = iter(value)

            # Insert all of the items from the value
            for subindex, subvalue in zip(range(start, stop, step), iterator):
                self[subindex] = subvalue

            # If the step is exactly 1, then we want to insert new values
            if step == 1:
                # Loop over remaining items and insert them
                for counter, subvalue in enumerate(iterator):
                    self.insert(stop + counter, subvalue)

            # Nothing more to do
            return

        # If index is negative, add the length to it
        if index < 0:
            index += len(self)

        # Fetch the identifier
        original_identifier = self._identifier_from_index(index)

        # Generate the item value and set it
        with self._create_identifier_from_value(value) as identifier:
            self._redis.lset(self._key, index, identifier)

        # Delete the original nested value
        self._delete_by_identifier(original_identifier)

    def __delitem__(self, index: typing.Union[int, slice]) -> None:
        # If a slice is provided, splice the list
        if isinstance(index, slice):
            # Delete all required items
            for counter, subindex in enumerate(range(*index.indices(len(self)))):
                del self[subindex - counter]

            # Nothing more to do
            return

        # If index is negative, add the length to it
        if index < 0:
            index += len(self)

        # Fetch the identifier
        identifier = self._identifier_from_index(index)

        # Check if index is the first item
        if index == 0:
            # Index 0 == lpop
            self._redis.lpop(self._key, 1)
        elif index == len(self) - 1:
            # Index -1 == rpop
            self._redis.rpop(self._key, 1)
        else:
            # Generate a temporary value
            temporary_value = os.urandom(64).hex()

            # Use a pipeline to execute all of the actions atomically
            pipeline = self._redis.pipeline()

            # Set the temporary value in the selected index
            pipeline.lset(self._key, index, temporary_value)

            # Remove 1 item with the temporary value
            pipeline.lrem(self._key, 1, temporary_value)

            # Execute all pipeline actions
            pipeline.execute()

        # Delete the original nested object
        self._delete_by_identifier(identifier)

    def __len__(self) -> int:
        # Fetch the list length
        length = self._redis.llen(self._key)

        # Make sure the length is an integer
        if not isinstance(length, int):
            raise TypeError(length)

        # Return the length
        return length

    def __repr__(self) -> str:
        # Format the data like a dictionary
        return "[%s]" % ", ".join(repr(item) for item in self)

    def __eq__(self, other: typing.Any) -> bool:
        # Make sure the other item is a sequence
        if not isinstance(other, Sequence):
            return False

        # Make sure lengths are the same
        if len(self) != len(other):
            return False

        # Loop and check all items
        for index in range(len(self)):
            # Compare items
            if self[index] != other[index]:
                return False

        # Items match
        return True

    def insert(self, index: int, value: typing.Any) -> None:
        # If index is negative, add the length to it
        if index < 0:
            index += len(self)

        # Check whether the index is 0
        if index == 0:
            # Insert 0 == lpush
            with self._create_identifier_from_value(value) as identifier:
                self._redis.lpush(self._key, identifier)
        elif index == len(self):
            # Insert -1 == rpush
            with self._create_identifier_from_value(value) as identifier:
                self._redis.rpush(self._key, identifier)
        else:
            # Fetch the identifier
            original_identifier = self._identifier_from_index(index)

            # Make sure the original identifier is a string
            if not isinstance(original_identifier, str):
                original_identifier = original_identifier.decode(self.ENCODING)

            # Create the new value already
            with self._create_identifier_from_value(value) as identifier:
                # Generate a temporary value
                temporary_value = os.urandom(64).hex()

                # Use a pipeline to execute all of the actions atomically
                pipeline = self._redis.pipeline()

                # Set the temporary value at the current index
                pipeline.lset(self._key, index, temporary_value)

                # Insert the new item before the temporary item
                pipeline.linsert(self._key, "before", temporary_value, identifier)

                # Restore the original value at the required index
                pipeline.lset(self._key, index + 1, original_identifier)

                # Execute all pipeline actions
                pipeline.execute()

    def copy(self) -> typing.List[typing.Any]:
        # Create initial bunch
        output = list()

        # Loop over keys
        for value in self:
            # Try copying the value
            with contextlib.suppress(AttributeError):
                value = value.copy()

            # Update the bunch
            output.append(value)

        # Return the created output
        return output


# Extend nested types with list
NESTED_TYPES.append(NestedType("list", List, list))



from test_utilities import my_redis, dictionary_type, list_type


def test_readme_example(my_redis):
    # Create my_dictionary
    my_dict = dictionary_type("test-dict", my_redis)
    my_dict.test_value = "Hello World"
    my_dict.numbers = [10, 20, 30]
    my_dict.ages = {
        "User 1": 10,
        "User 2": 20,
        "User 3": 30,
    }

    # Change a user age
    my_dict.ages["User 3"] = 40

    # Show the variable types
    assert type(my_dict.ages) == dictionary_type
    assert type(my_dict.numbers) == list_type

    # Show the entire my_dictionary
    assert my_dict == {'test_value': 'Hello World', 'numbers': [10, 20, 30], 'ages': {'User 1': 10, 'User 2': 20, 'User 3': 40}}



import pytest

from test_utilities import my_dictionary, list_type, dictionary_type


def test_write_read_has_delete(my_dictionary):
    # Make sure the my_dictionary does not have the item
    assert "Hello" not in my_dictionary

    # Write the Hello value
    my_dictionary["Hello"] = "World"

    # Read the Hello value
    assert my_dictionary["Hello"] == "World"

    # Make sure the my_dictionary has the Hello item
    assert "Hello" in my_dictionary

    # Delete the item
    del my_dictionary["Hello"]

    # Make sure the my_dictionary does not have the item
    assert "Hello" not in my_dictionary

    # Make sure the getter now raises
    with pytest.raises(KeyError):
        assert my_dictionary["Hello"] == "World"


def test_write_read_random_types(my_dictionary):
    # Write some random types
    my_dictionary["Tuple"] = (0, 1, 2, 3)
    my_dictionary["Bytes"] = b"AAAA\x00BBBB"
    my_dictionary["ByteArray"] = bytearray(range(100))

    # Make sure these types are valid now
    assert my_dictionary["Tuple"] == (0, 1, 2, 3)
    assert my_dictionary["Bytes"] == b"AAAA\x00BBBB"
    assert my_dictionary["ByteArray"] == bytearray(range(100))


def test_write_recursive(my_dictionary):
    # Write the Hello value
    my_dictionary["Hello"] = {"World": 42}

    # Read the Hello value
    assert my_dictionary["Hello"] == dict(World=42)

    # Make sure the Hello value is a my_dictionary
    assert isinstance(my_dictionary["Hello"], dictionary_type)

    # Check nested bunching
    my_dictionary.Hello.Test = {"Value": 90}
    assert my_dictionary.Hello.Test.Value == 90

    # Test nesting
    assert isinstance(my_dictionary.Hello.Test, dictionary_type)

    # Setdefault with empty dict
    my_dictionary.setdefault("Test", dict())

    # Make sure subdict is a my_dictionary
    assert isinstance(my_dictionary.Test, dictionary_type)

    # Check type of subdict
    assert type(my_dictionary.Test) == dictionary_type

    # Test setdefaults on subdictionary
    my_dictionary["Test"].setdefaults(a=1)


def test_write_multiple_recursive(my_dictionary):
    # Assign all of the recursive dicts
    my_dictionary["Hello"] = dict(Sub1=dict(Sub2=dict(Sub3=["AAA", "BBB"])))

    # Check item types
    assert isinstance(my_dictionary.Hello, dictionary_type)
    assert isinstance(my_dictionary.Hello.Sub1, dictionary_type)
    assert isinstance(my_dictionary.Hello.Sub1.Sub2, dictionary_type)
    assert isinstance(my_dictionary.Hello.Sub1.Sub2.Sub3, list_type)

    # Check item value
    assert my_dictionary.Hello.Sub1.Sub2.Sub3[0] == "AAA"
    assert my_dictionary.Hello.Sub1.Sub2.Sub3[1] == "BBB"


def test_reassign(my_dictionary):
    # Initialize a value
    my_dictionary["A"] = "B"
    my_dictionary["A"] = "C"

    # Make sure reassign succeeded
    assert my_dictionary["A"] == "C"

    # Re-assign a nested variable
    my_dictionary["A"] = {"B": "C"}
    my_dictionary["A"] = {"B": "D"}

    # Make sure reassign was successful
    assert my_dictionary["A"]["B"] == "D"


def test_list_reassign(my_dictionary):
    # Initialize a value
    my_dictionary["A"] = [2, 3, 4]
    my_dictionary["A"] = [1, 2, 3]

    # Make sure reassign succeeded
    assert my_dictionary["A"] == [1, 2, 3]

    # Initialize a value
    my_dictionary["B"] = []
    my_dictionary["B"] = [1, 2, 3]

    # Make sure reassign succeeded
    assert my_dictionary["B"] == [1, 2, 3]

    # Initialize a value
    my_dictionary["C"] = [1, 2, 3]
    my_dictionary["C"] = []

    # Make sure reassign succeeded
    assert my_dictionary["C"] == []


def test_list_reassign_bunch(my_dictionary):
    # Initialize a value
    my_dictionary.A = [2, 3, 4]
    my_dictionary.A = [1, 2, 3]

    # Make sure reassign succeeded
    assert my_dictionary.A == [1, 2, 3]

    # Initialize a value
    my_dictionary.B = []
    my_dictionary.B = [1, 2, 3]

    # Make sure reassign succeeded
    assert my_dictionary.B == [1, 2, 3]

    # Initialize a value
    my_dictionary.C = [1, 2, 3]
    my_dictionary.C = []

    # Make sure reassign succeeded
    assert my_dictionary.C == []


def test_len(my_dictionary):
    # Make sure my_dictionary is empty
    assert not my_dictionary

    # Load value to my_dictionary
    my_dictionary["Hello"] = "World"

    # Make sure my_dictionary is not empty
    assert my_dictionary


def test_pop(my_dictionary):
    # Load value to my_dictionary
    my_dictionary["Hello"] = "World"

    # Pop the item from the my_dictionary
    assert my_dictionary.pop("Hello") == "World"

    # Make sure the my_dictionary is empty
    assert not my_dictionary


def test_popitem(my_dictionary):
    # Load value to my_dictionary
    my_dictionary["Hello"] = "World"

    # Pop the item from the my_dictionary
    assert my_dictionary.popitem() == ("Hello", "World")

    # Make sure the my_dictionary is empty
    assert not my_dictionary


def test_copy(my_dictionary):
    # Load values to my_dictionary
    my_dictionary["Hello1"] = "World1"
    my_dictionary["Hello2"] = "World2"

    # Copy the my_dictionary and compare
    copy = my_dictionary.copy()

    # Check copy
    assert isinstance(copy, dict)
    assert copy == {"Hello1": "World1", "Hello2": "World2"}


def test_equals(my_dictionary):
    # Load values to my_dictionary
    my_dictionary["Hello1"] = "World1"
    my_dictionary["Hello2"] = "World2"

    assert my_dictionary == {"Hello1": "World1", "Hello2": "World2"}
    assert my_dictionary != {"Hello1": "World1", "Hello2": "World2", "Hello3": "World3"}
    assert my_dictionary != {"Hello2": "World2", "Hello3": "World3"}


def test_representation(my_dictionary):
    # Make sure looks good empty
    assert repr(my_dictionary) == "{}"

    # Load some values
    my_dictionary["Hello"] = "World"
    my_dictionary["Other"] = {"Test": 1}

    # Make sure looks good with data
    assert repr(my_dictionary) in ["{'Hello': 'World', 'Other': {'Test': 1}}", "{'Other': {'Test': 1}, 'Hello': 'World'}"] + ["{u'Hello': u'World', u'Other': {u'Test': 1}}", "{u'Other': {u'Test': 1}, u'Hello': u'World'}"]


def test_delete(my_dictionary):
    # Load some values
    my_dictionary["Persistent"] = "Test"
    my_dictionary["Volatile"] = "Forbidden"

    # Make sure values exist
    assert "Persistent" in my_dictionary
    assert "Volatile" in my_dictionary

    # Compare values
    assert my_dictionary["Persistent"] == "Test"
    assert my_dictionary["Volatile"] == "Forbidden"

    # Delete one value
    del my_dictionary["Volatile"]

    # Make sure persistent value exists
    assert "Persistent" in my_dictionary
    assert my_dictionary["Persistent"] == "Test"

    # Try deleting a non-existent value
    with pytest.raises(KeyError):
        del my_dictionary["Non-Existent"]


def test_clear(my_dictionary):
    # Load some values
    my_dictionary["Hello"] = "World"
    my_dictionary["Other"] = {"Test": 1}

    # Fetch other my_dictionary
    other = my_dictionary["Other"]

    # Make sure other is not empty
    assert other

    # Clear the my_dictionary
    my_dictionary.clear()

    # Make sure my_dictionary is empty
    assert not my_dictionary

    # Make sure other does not exist
    assert not other


def test_setdefaults(my_dictionary):
    # Set some values
    my_dictionary.setdefaults({"a": "b"}, c="d")

    # Check my_dictionary structure
    assert my_dictionary == {"a": "b", "c": "d"}


def test_update(my_dictionary):
    # Set some values
    my_dictionary.update({"a": "b"}, c="d")

    # Check my_dictionary structure
    assert my_dictionary == {"a": "b", "c": "d"}


def test_bunch_mode(my_dictionary):
    # Assign values
    my_dictionary.test_value = 10

    # Make sure the item was written
    assert my_dictionary["test_value"] == 10

    # Set another value
    my_dictionary["another"] = 5

    # Try getting the value
    assert my_dictionary.another == 5

    # Try deleting the value
    del my_dictionary.another

    # Check final my_dictionary values
    assert my_dictionary == {"test_value": 10}



import pytest

from test_utilities import my_dictionary, my_list, list_type, dictionary_type


def test_create(my_dictionary):
    # Create the list type
    my_dictionary["Test"] = [1, 2, 3]

    # Make sure list is created
    assert isinstance(my_dictionary["Test"], list_type)


def test_nested_arrays(my_list):
    # Set subarray
    my_list.append([0, 1, 2, 3])

    # Make sure list is created
    assert isinstance(my_list[0], list_type)


def test_append(my_list):
    # Set my_list items
    my_list.append(1)
    my_list.append(2)
    my_list.append(3)

    # Make sure all items are working properly
    assert my_list[0] == 1
    assert my_list[1] == 2
    assert my_list[2] == 3


def test_contains(my_list):
    # Set my_list items
    my_list.append(1)

    # Make sure all items are working properly
    assert 1 in my_list
    assert 4 not in my_list


def test_delete(my_list):
    # Set my_list items
    my_list.append(1)

    # Delete from list
    assert 1 in my_list
    del my_list[0]
    assert 1 not in my_list

    # Delete non-existent item
    with pytest.raises(IndexError):
        del my_list[0]


def test_pop(my_list):
    # Set my_list items
    my_list.append(1)
    my_list.append(2)
    my_list.append(3)
    my_list.append(4)

    # Delete from list
    assert my_list.pop() == 4
    assert my_list.pop() == 3
    assert my_list.pop(0) == 1
    assert len(my_list) == 1

    # Add new items to my_list
    my_list.append(90)
    my_list.append(91)
    my_list.append(92)
    my_list.append(93)

    # Make sure poping works
    assert my_list.pop(-1) == 93
    assert my_list.pop() == 92


def test_insert(my_list):
    # Set my_list items
    my_list.append(1)
    my_list.append(2)
    my_list.append(3)
    my_list.append(4)

    # Insert in list
    my_list.insert(0, 9)
    assert my_list[0] == 9

    # Insert in list
    my_list.insert(1, 8)
    assert my_list[1] == 8

    # Test insert 0
    my_list.insert(0, 100)
    assert my_list[0] == 100

    # Test insert -1
    my_list.insert(-1, 99)
    assert my_list[-2] == 99
    assert my_list[len(my_list) - 2] == 99

    # Test insert len
    my_list.insert(len(my_list), 200)
    assert my_list[-1] == 200


def test_slice(my_list):
    # Insert to list
    for index in range(10):
        my_list.append(index)

    # Check slice
    assert my_list[3:6] == [3, 4, 5]

    # Delete slice
    del my_list[2:5]

    # Check slice
    assert my_list == [0, 1, 5, 6, 7, 8, 9]

    # Set slice
    my_list[3:6] = range(6)

    # Check slice
    assert my_list == [0, 1, 5, 0, 1, 2, 3, 4, 5, 9]

    # Get extended slices
    assert my_list[3:8:2] == [0, 2, 4]

    # Delete extended slices
    del my_list[3:8:2]

    # Check deleted slice
    assert my_list == [0, 1, 5, 1, 3, 5, 9]

    # Insert extended slice (should fail)
    with pytest.raises(ValueError):
        my_list[2:9:3] = range(10)

    # Insert extended slice
    my_list[2:9:3] = range(90, 92)

    # Make sure the my_list is as expected
    assert my_list == [0, 1, 90, 1, 3, 91, 9]


def test_negative_index(my_list):
    # Insert to list
    my_list.append(1)
    my_list.append(2)

    # Make sure the negative index works
    assert my_list[-1] == 2
    assert my_list[-2] == 1


def test_equals(my_list):
    # Insert to list
    my_list.append(1)
    my_list.append(2)
    my_list.append(3)

    # Compare
    assert my_list != [2, 3]
    assert my_list != None
    assert my_list != [2, 3, 4]
    assert my_list == [1, 2, 3]

    # This is a convenient feature, regular lists don't do this.
    assert my_list == (1, 2, 3)


def test_assignment(my_list):
    # Insert to list
    my_list.append(1)
    my_list.append(2)

    # Assign item
    my_list[1] = 8

    # Check assignment
    assert my_list == [1, 8]


def test_dict_assignment(my_list):
    # Insert to list
    my_list.append(dict(a=["Hello", "World"]))

    # Fetch the dict
    assert isinstance(my_list[0], dictionary_type)

    # Make sure the value is set correctly
    assert my_list[0].a == ["Hello", "World"]


def test_reassign(my_list):
    # Create a new sublist
    my_list.append([1, 2, 3])

    # Make sure the item is a list_type
    assert isinstance(my_list[0], list_type)

    # Make sure the item has the value we expect
    assert my_list[0] == [1, 2, 3]

    # Reassign the value
    my_list[0] = [2, 3, 4]

    # Make sure the item is a list_type

    # Make sure the item has the value we expect
    assert my_list[0] == [2, 3, 4]

    # Reassign the value
    my_list[0] = []

    # Make sure the item is a list_type
    assert isinstance(my_list[0], type(my_list))

    # Make sure the item has the value we expect
    assert my_list[0] == []

    # Reassign the value
    my_list[0] = [2, 3, 4]

    # Make sure the item is a list_type
    assert isinstance(my_list[0], type(my_list))

    # Make sure the item has the value we expect
    assert my_list[0] == [2, 3, 4]



import json
import pytest

from rednest import *

from test_utilities import my_dictionary, my_list


def test_dict_encoding(my_dictionary):
    # Assign values
    my_dictionary["AAA"] = 5
    my_dictionary["AAB"] = [1, 2, 3]

    # Dump my_dictionary
    assert json.dumps(my_dictionary)


def test_array_encoding(my_list):
    # Insert to list
    for x in range(10):
        my_list.append(x)

    # Encode list
    assert json.dumps(my_list) == '[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]'



# Import redis encoder
from rednest.encoder import Encoder

# Import nested objects
from rednest.list import List
from rednest.dictionary import Dictionary

# Import base objects for extendability
from rednest.nested import Nested, NestedType, NESTED_TYPES


import os
import redis
import pytest

# Rednest testing types
from rednest import List as list_type, Dictionary as dictionary_type

# All supported test connections
REDIS_CONNECTIONS = [redis.Redis(), redis.Redis(decode_responses=True)]


@pytest.fixture(params=REDIS_CONNECTIONS)
def my_redis(request):
    return request.param


@pytest.fixture(params=REDIS_CONNECTIONS)
def my_dictionary(request):
    # Generate random name
    rand_name = os.urandom(4).hex()

    # Create a random my_dictionary
    return dictionary_type(rand_name, redis=request.param)


@pytest.fixture(params=REDIS_CONNECTIONS)
def my_list(request):
    # Generate random name
    rand_name = os.urandom(4).hex()

    # Create a random my_dictionary
    return list_type(rand_name, redis=request.param)
